from .rainbow import *
import batman as batman


class SimulatedRainbow(Rainbow):
    def __init__(
        self,
        signal_to_noise=100,
        tlim=[-2.5, 2.5] * u.hour,
        dt=10 * u.minute,
        time=None,
        wlim=[0.5, 5] * u.micron,
        R=20,
        dw=None,
        wavelength=None,
        star_flux=None,
    ):
        """
        Create a simulated rainbow object.

        Parameters
        ----------

        signal_to_noise : float
            The signal-to-noise per wavelength per time.
            For example, S/N=100 would mean that the
            uncertainty on the flux for each each
            wavelength-time data point will be 1%.

        tlim : list or array of astropy.units.Quantity
            The pip install -e '.[develop]'[min, max] times for creating the time grid.
            These should have astropy units of time.
        dt : astropy.units.Quantity
            The d(time) bin size for creating a grid
            that is uniform in linear space.
        time : array of astropy.units.Quantity
            An array of times, if you just want to give
            it an entirely custom array.

        The time-setting order of precendence is:
            1) time
            2) tlim + dt

        wlim : list or array of astropy.units.Quantity
            The [min, max] wavelengths for creating the grid.
            These should have astropy units of wavelength.
        R : float
            The spectral resolution for creating a grid
            that is uniform in logarithmic space.
        dw : astropy.units.Quantity
            The d(wavelength) bin size for creating a grid
            that is uniform in linear space.
        wavelength : array of astropy.units.Quantity
            An array of wavelengths, if you just want to give
            it an entirely custom array.

        The wavelength-setting order of precendence is:
            1) wavelength
            2) wlim + dw
            3) wlim + R

        star_flux : numpy 1D array
            An array of fluxes corresponding to the supplied wavelengths.
            If left blank, the code assumes a normalized flux of
            flux(wavelength) = 1 for all wavelengths.
        planet : boolean
            Set to True to add a transiting planet.
        planet_params: Dictionary
            Stores planetary parameters to model, read description for
            Rainbow.add_planet transit() for more details.
        planet_radius: float or 1D numpy array
            Planet Radius/Stellar Radius of planet.  Use 1D if wavelength-
            dependent.  Must be same shape as wavelength array.
        """
        Rainbow.__init__(self)

        # set up the wavelength grid
        self._setup_fake_wavelength_grid(wlim=wlim, R=R, dw=dw, wavelength=wavelength)

        # set up the time grid
        self._setup_fake_time_grid(tlim=tlim, dt=dt, time=time)

        # Save SNR.
        self.metadata["signal_to_noise"] = signal_to_noise

        # If the flux of the star is not given,
        # assume a continuum-normlized flux where fx=1 at all wavelengths.
        if star_flux is None:
            self.fluxlike["model"] = np.ones(self.shape)

        # If the flux vs wavelength of the star is supplied,
        # include it in the model.
        else:
            # Check to make sure the flux and wavelengths
            # have the same shape.
            if len(star_flux) == len(self.wavelike["wavelength"]):
                self.fluxlike["model"] = np.transpose([star_flux] * self.shape[1])

        # Set uncertainty.
        self.fluxlike["uncertainty"] = self.fluxlike["model"] / signal_to_noise
        self.fluxlike["flux"] = np.random.normal(
            self.fluxlike["model"], self.fluxlike["uncertainty"]
        )

    def _setup_fake_time_grid(
        self, tlim=[-2.5 * u.hour, 2.5 * u.hour], dt=1 * u.minute, time=None
    ):
        """
        Create a fake time grid.

        Parameters
        ----------

        tlim : list or array of astropy.units.Quantity
            The [min, max] times for creating the time grid.
            These should have astropy units of time.
        dt : astropy.units.Quantity
            The d(time) bin size for creating a grid
            that is uniform in linear space.
        time : array of astropy.units.Quantity
            An array of times, if you just want to give
            it an entirely custom array.

        The time-setting order of precendence is:
            1) time
            2) tlim + dt
        """
        # check we're trying to do exactly one thing
        if (tlim is None) and (time is None):
            raise RuntimeError("Please specify either `tlim` or `time`.")

        if time is None:
            t_unit = tlim[0].unit
            t_unit.to("s")
            time = np.arange(tlim[0] / t_unit, tlim[1] / t_unit, dt / t_unit) * t_unit
        else:
            t_unit = time.unit

        self.timelike["time"] = u.Quantity(time)
        # TODO, make this match up better with astropy time

    def _setup_fake_wavelength_grid(
        self, wlim=[0.5 * u.micron, 5 * u.micron], R=100, dw=None, wavelength=None
    ):
        """
        Create a fake wavelength grid.

        Parameters
        ----------

        wlim : list or array of astropy.units.Quantity
            The [min, max] wavelengths for creating the grid.
            These should have astropy units of wavelength.
        R : float
            The spectral resolution for creating a grid
            that is uniform in logarithmic space.
        dw : astropy.units.Quantity
            The d(wavelength) bin size for creating a grid
            that is uniform in linear space.
        wavelength : array of astropy.units.Quantity
            An array of wavelengths, if you just want to give
            it an entirely custom array.

        The wavelength-setting order of precendence is:
            1) wavelength
            2) wlim + dw
            3) wlim + R
        """

        # check we're trying to do exactly one thing
        if (wlim is None) and (wavelength is None):
            raise RuntimeError("Please specify either `wlim` or `wavelength`.")

        # create a linear or logarithmic grid
        if wavelength is None:
            # check that we're
            if (R is None) and (dw is None):
                raise RuntimeError("Please specify either `R` or `dw`.")

            w_unit = wlim[0].unit
            if dw is None:
                self.metadata["R"] = R
                # self.metadata["wscale"] = "log"

                logw_min = np.log(wlim[0] / w_unit)
                logw_max = np.log(wlim[1] / w_unit)
                logw = np.arange(logw_min, logw_max, 1 / R)
                wavelength = np.exp(logw) * w_unit

            elif dw is not None:
                self.metadata["dw"] = dw
                # self.metadata["wscale"] = "linear"
                wavelength = (
                    np.arange(wlim[0] / w_unit, wlim[1] / w_unit, self.dw / w_unit)
                    * w_unit
                )

        # or just make sure the wavelength grid has units
        elif wavelength is not None:
            w_unit = wavelength.unit

        # make sure the wavelength array has units
        self.wavelike["wavelength"] = u.Quantity(wavelength)
        self._guess_wscale()

        # this should break if the units aren't length
        w_unit.to("m")

    def inject_transit(self, planet_params={}, planet_radius=0.1):

        """
        Simulate a wavelength-dependent planetary transit using
        batman.

        Parameters
        ----------

        planet_params : Dictionary
            Values for planetary parameters for use in batman modelling.
            Any values not supplied will be set to defaults:
                "t0" = time of inferior conjunction (days) (default 0)
                "per" = orbital period (days) (detault 1)
                "a" = semi-major axis (units of stellar radii) (default 15)
                "inc" = inclination (degrees) (default 90)
                "ecc" = eccentricity (default 0)
                "w" = longitude of periastron (degrees)(default 0)
                "limb_dark" = limb-darkening model (default "nonlinear"), possible
                    values described in more detail in batman documentation
                "u" = limb-darkening coefficients (default [0.5, 0.1, 0.1, -0.1])
                    Can take 3 forms:
                    -A single value (if limb-darkening law requires only one value)
                    -A 1D list/array of coefficients corresponding to the limb-darkening
                    law
                    -A 2D array of the form (n_wavelengths, n_coefficients) where
                    each row is the set of limb-darkening coefficients corresponding
                    to a single wavelength

                    Note that this currently does not calculate the appropriate
                    coefficient vs wavelength variations itself- there exist codes
                    (such as hpparvi/PyLDTk and nespinoza/limb-darkening) which
                    can be used for this.

            example value: planet_params = {"a":12, "inc":87}

        planet_radius = Two options:
                1D array with same dimensions as wavelength array,
                    each value corresponds to planet radius/stellar radius at that
                    wavelength.
                float representing Rp/Rstar if the radius is not wavelength-dependent.
            example value: planet_radius = 0.01,

        """

        # First, make sure planet_radius has the right dimension.
        if type(planet_radius) != float and len(planet_radius) != self.nwave:
            print(
                "Invalid planet radius array: must be float or have shape "
                + str(np.shape(self.wavelike["wavelength"]))
            )

        # Defaults for planet simulation.
        defaults = {
            "t0": 0,
            "per": 3,
            "a": 10,
            "inc": 90,
            "ecc": 0,
            "w": 0,
            "limb_dark": "nonlinear",
            "u": [0.5, 0.1, 0.1, -0.1],
        }

        # Read in planet parameters.
        for i in range(len(planet_params.keys())):
            key = list(planet_params.keys())[i]
            if key in list(defaults.keys()):
                defaults[key] = planet_params[key]
            else:
                print("Warning: " + str(key) + " not a valid parameter")

        # Initialize batman model.
        params = batman.TransitParams()
        params.t0 = defaults["t0"]
        params.per = defaults["per"]
        params.a = defaults["a"]
        params.inc = defaults["inc"]
        params.ecc = defaults["ecc"]
        params.w = defaults["w"]
        params.limb_dark = defaults["limb_dark"]

        # Deal with limb-darkening.
        if len(np.shape(defaults["u"])) < 2:  # Coefficients constant with wavelength
            u_arr = np.tile(defaults["u"], (self.nwave, 1))

        elif (
            len(np.shape(defaults["u"])) == 2
        ):  # 2D array of coefficients, along wavelength axis
            if np.shape(defaults["u"])[0] != self.nwave:
                print("Shape of limb-darkening array does not match wavelengths.")
                return
            u_arr = defaults["u"]

        else:
            print("Invalid limb-darkening coefficient array.")

        # Read in planetary radius.
        if type(planet_radius) == float:
            rprs = np.zeros(self.nwave) + planet_radius
        else:
            rprs = planet_radius

        planet_flux = np.zeros((self.nwave, self.ntime))

        for i in range(self.nwave):
            params.rp = rprs[i]
            params.u = u_arr[i]
            # print(params.u)
            try:
                m
            except NameError:
                m = batman.TransitModel(params, self.timelike["time"].to("day").value)
            planet_flux[i] = m.light_curve(params)

        result = self._create_copy()
        result.fluxlike["model"] *= planet_flux
        result.fluxlike["flux"] *= planet_flux

        return result



from .stsci import *
from .eureka import *
from .rainbownpy import *
from .nestor_niriss_numpy import *
from .text import *

# construct a dictionary of available readers
available_readers = {k: globals()[k] for k in globals() if k[0:5] == "from_"}


def guess_reader(filepath, format=None):
    """
    A wrapper to guess the appropriate reader from the filename
    (and possibily an explicitly-set file format string).

    Parameters
    ----------
    filepath : str
        The path to the file, or a glob-friendly string
        pointing to a group of files that should all be
        loaded together (for example, if an exposure was
        split into multiple segments).
    format : str, None
        The file format to use.
    """
    import fnmatch, glob
    from ...imports import expand_filenames

    # get all the possible filenames (= expand wildcard)
    filenames = expand_filenames(filepath)

    # if format='abcdefgh', return the `from_abcdefgh` function
    if format is not None:
        return available_readers[f"from_{format}"]
    # does it look like a .rainbow.npy chromatic file?
    elif fnmatch.fnmatch(filenames[0], "*.rainbow.npy"):
        return from_rainbownpy
    # does it look like a .rainbow.npy chromatic file?
    elif fnmatch.fnmatch(filenames[0], "*order*.npy"):
        return from_nestor_niriss_numpy
    # does it look like a STScI x1dints.fits file?
    elif fnmatch.fnmatch(filenames[0], "*x1dints.fits") or fnmatch.fnmatch(
        filenames[0], "*extract_1d.fits"
    ):
        return from_x1dints
    # does it look like an Eureka! save file?
    elif (
        fnmatch.fnmatch(filenames[0], "*S3_*_Save.dat")
        or fnmatch.fnmatch(filenames[0], "*S3_*_Save.h5")
        or fnmatch.fnmatch(filenames[0], "*S3_*_Save.txt")
    ):
        return from_eureka
    elif fnmatch.fnmatch(filenames[0], "*.txt") or fnmatch.fnmatch(
        filenames[0], "*.csv"
    ):
        return from_text
    else:
        raise RuntimeError(f"🌈 Failed to guess a good reader for {filenames}.")



from ..rainbows import *
from .setup_tests import *


def test_imshow():
    plt.figure()
    SimulatedRainbow(R=10).imshow()

    plt.figure()
    SimulatedRainbow(dw=0.2 * u.micron).imshow()

    plt.figure()
    Rainbow(
        wavelength=np.arange(1, 5) * u.micron,
        time=np.arange(1, 6) * u.hour,
        flux=np.ones((4, 5)),
    ).imshow()

    fi, ax = plt.subplots(2, 1, sharex=True)
    SimulatedRainbow(R=10).imshow(w_unit="nm", ax=ax[0])
    SimulatedRainbow(dw=0.2 * u.micron).imshow(ax=ax[1], w_unit="nm")
    plt.savefig(os.path.join(test_directory, "imshow-demonstration.pdf"))


def test_plot():
    SimulatedRainbow(R=10).plot()
    plt.savefig(os.path.join(test_directory, "plot-demonstration.pdf"))


def test_animate():
    # test a transit, since along both dimensions
    d = SimulatedRainbow(dw=0.1 * u.micron, dt=5 * u.minute, signal_to_noise=1000)
    theta = np.linspace(0, 2 * np.pi, d.nwave)
    planet_radius = np.sin(theta) * 0.05 + 0.15
    e = d.inject_transit(planet_radius=planet_radius)
    scatterkw = dict()
    e.animate_lightcurves(
        filename=os.path.join(test_directory, "animate-lightcurves-demonstration.gif"),
        scatterkw=scatterkw,
    )
    e.animate_spectra(
        filename=os.path.join(test_directory, "animate-spectra-demonstration.gif"),
        scatterkw=scatterkw,
    )


def test_wavelength_cmap():

    r = SimulatedRainbow(R=10)

    # can we set up the wavelength-based color map
    r._setup_wavelength_colors(cmap=one2another("black", "red"))

    # test a few examples
    assert r.get_wavelength_color(r.wavelength[0]) == (0.0, 0.0, 0.0, 1.0)
    assert r.get_wavelength_color(r.wavelength[-1]) == (1.0, 0.0, 0.0, 1.0)



from ..imports import *
from ..talker import Talker
from .readers import *
from .writers import *


class Rainbow(Talker):
    """
    Rainbow objects represent the flux of some astrophysical
    object as a function of both wavelength and time.

    The general stucture of a Rainbow object contains a
    1D array of wavelengths, a 1D array of times, and a
    2D array of flux values (with row as wavelength and
    column as time).
    """

    # all Rainbows must contain these core dictionaries
    _core_dictionaries = ["fluxlike", "timelike", "wavelike", "metadata"]

    # define which axis is which
    waveaxis = 0
    timeaxis = 1

    def __init__(
        self,
        filepath=None,
        format=None,
        wavelength=None,
        time=None,
        flux=None,
        uncertainty=None,
        wavelike=None,
        timelike=None,
        fluxlike=None,
        metadata=None,
        **kw,
    ):
        """
        Initialize a Rainbow object from a file, from arrays
        with appropriate units, or from dictionaries with
        appropriate ingredients.

        Parameters
        ----------

        filepath : str
            The filepath pointing to the file or group of files
            that should be read.
        format : str
            The file format of the file to be read. If None,
            the format will be guessed automatically from the
            filepath.

        wavelength : astropy.unit.Quantity
            A 1D array of wavelengths, in any unit.
        time : astropy.unit.Quantity or astropy.unit.Time
            A 1D array of times, in any unit.
        flux : np.array
            A 2D array of flux values.
        uncertainty : np.array
            A 2D array of uncertainties, associated with the flux.

        wavelike : dict
            A dictionary containing 1D arrays with the same
            shape as the wavelength axis. It must at least
            contain the key 'wavelength', which should have
            astropy units of wavelength associated with it.
        timelike : dict
            A dictionary containing 1D arrays with the same
            shape as the time axis. It must at least
            contain the key 'time', which should have
            astropy units of time associated with it.
        fluxlike : dict
            A dictionary containing 2D arrays with the shape
            of (nwave, ntime), like flux. It must at least
            contain the key 'flux'.
        metadata : dict
            A dictionary containing all other metadata
            associated with the dataset, generally lots of
            individual parameters or comments.

        Examples
        --------
        ```
        # initalize from a file
        r1 = Rainbow('my-neat-file.abc', format='abcdefgh')

        # initalize from arrays
        r2 = Rainbow(wavelength=np.linspace(1, 5, 50)*u.micron,
                     time=np.linspace(-0.5, 0.5, 100)*u.day,
                     flux=np.random.normal(0, 1, (50, 100)))

        # initialize from dictionaries
        f3 = Rainbow(wavelike=dict(wavelength=np.linspace(1, 5, 50)*u.micron),
                     timelike=dict(time=np.linspace(-0.5, 0.5, 100)*u.day),
                     fluxlike=dict(flux=np.random.normal(0, 1, (50, 100))))

        """
        # metadata are arbitrary types of information we need
        self.metadata = {}

        # wavelike quanities are 1D arrays with nwave elements
        self.wavelike = {}

        # timelike quantities are 1D arrays with ntime elements
        self.timelike = {}

        # fluxlike quantities are 2D arrays with nwave x time elements
        self.fluxlike = {}

        # try to intialize from the exact dictionaries needed
        if (
            (type(wavelike) == dict)
            and (type(timelike) == dict)
            and (type(fluxlike) == dict)
        ):
            self._initialize_from_dictionaries(
                wavelike=wavelike,
                timelike=timelike,
                fluxlike=fluxlike,
                metadata=metadata,
            )
        # then try to initialize from arrays
        elif (wavelength is not None) and (time is not None) and (flux is not None):
            self._initialize_from_arrays(
                wavelength=wavelength,
                time=time,
                flux=flux,
                uncertainty=uncertainty,
                **kw,
            )
        # then try to initialize from a file
        elif (type(filepath) == str) or (type(filepath) == list):
            self._initialize_from_file(filepath=filepath, format=format)

        # finally, tidy up by guessing the wavelength scale
        self._guess_wscale()

    def save(self, filepath="test.rainbow.npy", format=None, **kw):
        """
        Save this Rainbow out to a file.

        Parameters
        ----------
        filepath : str
            The filepath pointing to the file to be written.
            (For now, it needs a `.rainbow.npy` extension.)
        format : str
            The file format of the file to be written. If None,
            the format will be guessed automatically from the
            filepath."""

        # figure out the best writer
        writer = guess_writer(filepath, format=format)

        # use that writer to save the file
        writer(self, filepath, **kw)

    def _initialize_from_dictionaries(
        self, wavelike={}, timelike={}, fluxlike={}, metadata={}
    ):
        """
        Populate from dictionaries in the correct format.

        Parameters
        ----------
        wavelike : dict
            A dictionary containing 1D arrays with the same
            shape as the wavelength axis. It must at least
            contain the key 'wavelength', which should have
            astropy units of wavelength associated with it.
        timelike : dict
            A dictionary containing 1D arrays with the same
            shape as the time axis. It must at least
            contain the key 'time', which should have
            astropy units of time associated with it.
        fluxlike : dict
            A dictionary containing 2D arrays with the shape
            of (nwave, ntime), like flux. It must at least
            contain the key 'flux'.
        metadata : dict
            A dictionary containing all other metadata
            associated with the dataset, generally lots of
            individual parameters or comments.
        """

        # update the four core dictionaries
        self.wavelike.update(**wavelike)
        self.timelike.update(**timelike)
        self.fluxlike.update(**fluxlike)
        self.metadata.update(**metadata)

        # validate that something reasonable got populated
        self._validate_core_dictionaries()

    def _get_core_dictionaries(self):
        """
        Get the core dictionaries of this Rainbow.

        Returns
        -------
        core : dict
            Dictionary containing the keys
            ['wavelike', 'timelike', 'fluxlike', 'metadata']
        """
        return {k: vars(self)[k] for k in self._core_dictionaries}

    def _initialize_from_arrays(
        self, wavelength=None, time=None, flux=None, uncertainty=None, **kw
    ):
        """
        Populate from arrays.

        Parameters
        ----------
        wavelength : astropy.unit.Quantity
            A 1D array of wavelengths, in any unit.
        time : astropy.unit.Quantity or astropy.unit.Time
            A 1D array of times, in any unit.
        flux : np.array
            A 2D array of flux values.
        uncertainty : np.array
            A 2D array of uncertainties, associated with the flux.
        """

        # store the wavelength
        self.wavelike["wavelength"] = wavelength

        # store the time
        self.timelike["time"] = time

        # store the flux and uncertainty
        self.fluxlike["flux"] = flux
        self.fluxlike["uncertainty"] = uncertainty

        # validate that something reasonable got populated
        self._validate_core_dictionaries()

    def _initialize_from_file(self, filepath=None, format=None):
        """
        Populate from a filename or group of files.

        Parameters
        ----------

        filepath : str
            The filepath pointing to the file or group of files
            that should be read.
        format : str
            The file format of the file to be read. If None,
            the format will be guessed automatically from the
            filepath.
        """

        # make sure we're dealing with a real filename
        assert filepath is not None

        # pick the appropriate reader
        reader = guess_reader(filepath=filepath, format=format)
        reader(self, filepath)

        # validate that something reasonable got populated
        self._validate_core_dictionaries()

    def _create_copy(self):
        """
        Create a copy of self, with the core dictionaries copied.
        """
        new = type(self)()
        new._initialize_from_dictionaries(
            **copy.deepcopy(self._get_core_dictionaries())
        )
        return new

    def _guess_wscale(self, relative_tolerance=0.01):
        """
        Try to guess the wscale from the wavelengths.

        Parameters
        ----------

        relative_tolerance : float
            The fractional difference to which the differences
            between wavelengths should match in order for a
            linear or logarithmic wavelength scale to be
            assigned. For example, the default value of 0.01
            means that the differences between all wavelength
            must be within 1% of each other for the wavelength
            scale to be called linear.
        """

        # give up if there's no wavelength array
        if self.wavelength is None:
            return "?"

        # calculate difference arrays
        w = self.wavelength.value
        dw = np.diff(w)
        dlogw = np.diff(np.log(w))

        # test the three options
        if np.allclose(dw, np.median(dw), rtol=relative_tolerance):
            self.metadata["wscale"] = "linear"
        elif np.allclose(dlogw, np.median(dlogw), rtol=relative_tolerance):
            self.metadata["wscale"] = "log"
        else:
            self.metadata["wscale"] = "?"

    def _guess_tscale(self, relative_tolerance=0.01):
        """
        Try to guess the tscale from the times.

        Parameters
        ----------

        relative_tolerance : float
            The fractional difference to which the differences
            between times should match in order for us to call
            the times effectively uniform, or for us to treat
            them more carefully as an irregular or gappy grid.
        """

        # give up if there's no time array
        if self.time is None:
            return "?"

        # calculate difference arrays
        t = self.time.value
        dt = np.diff(t)

        # test the three options
        if np.allclose(dt, np.median(dt), rtol=relative_tolerance):
            self.metadata["tscale"] = "uniform"
        else:
            self.metadata["tscale"] = "?"

    @property
    def wavelength(self):
        """
        The 1D array of wavelengths (with astropy units of length).
        """
        return self.wavelike.get("wavelength", None)

    @property
    def time(self):
        """
        The 1D array of wavelengths (with astropy units of length).
        """
        return self.timelike.get("time", None)

    @property
    def flux(self):
        """
        The 2D array of fluxes (row = wavelength, col = time).
        """
        return self.fluxlike.get("flux", None)

    @property
    def uncertainty(self):
        """
        The 2D array of uncertainties on the fluxes.
        """
        return self.fluxlike.get("uncertainty", None)

    def __getattr__(self, key):
        """
        If an attribute/method isn't explicitly defined,
        try to pull it from one of the core dictionaries.

        Let's say you want to get the 2D uncertainty array
        but don't want to type `self.fluxlike['uncertainty']`.
        You could instead type `self.uncertainty`, and this
        would try to search through the four standard
        dictionaries to pull out the first `uncertainty`
        it finds.

        Parameters
        ----------
        key : str
            The attribute we're trying to get.
        """
        if key not in self._core_dictionaries:
            for dictionary_name in self._core_dictionaries:
                try:
                    return self.__dict__[dictionary_name][key]
                except KeyError:
                    pass
        message = f"🌈.{key} does not exist for this Rainbow"
        raise AttributeError(message)

    # TODO - what should we do with __setattr__?
    #   actually allow to reset things in metadata?
    #   give a warning that you try to set something you shouldn't?
    #   if things have the right size, just organize them

    @property
    def _nametag(self):
        """
        This short phrase will preface everything
        said with `self.speak()`.
        """
        return f"🌈({self.nwave}w, {self.ntime}t)"

    @property
    def shape(self):
        """
        The shape of the flux array (nwave, ntime).
        """
        return (self.nwave, self.ntime)

    @property
    def nwave(self):
        """
        The number of wavelengths.
        """
        if self.wavelength is None:
            return 0
        else:
            return len(self.wavelength)

    @property
    def ntime(self):
        """
        The number of times.
        """
        if self.time is None:
            return 0
        else:
            return len(self.time)

    @property
    def nflux(self):
        """
        The total number of fluxes.
        """
        return np.prod(self.shape)

    def _validate_core_dictionaries(self):
        """
        Do some simple checks to make sure this Rainbow
        is populated with the minimal data needed to do anything.
        It shouldn't be run before the Rainbow is fully
        initialized; otherwise, it might complain about
        a half-populated object.
        """

        # make sure there are some times + wavelengths defined
        if self.ntime is None:
            warnings.warn(
                f"""
            No times are defined for this Rainbow.
            """
            )
        if self.nwave is None:
            warnings.warn(
                f"""
            No wavelengths are defined for this Rainbow.
            """
            )

        # does the flux have the right shape?
        if self.shape != self.flux.shape:
            message = "Flux array shape does not match (wavelength, time)."
            if self.shape == self.flux.shape[::-1]:
                warnings.warn(
                    f"""
                {message}
                Any chance it's transposed?"""
                )
            else:
                warnings.warn(
                    f"""
                {message}"""
                )

    def __getitem__(self, key):
        """
        Trim a rainbow by indexing, slicing, or masking.
        Two indices must be provided (`[:,:]`).

        Examples
        --------
        ```
        r[:,:]
        r[10:20, :]
        r[np.arange(10,20), :]
        r[r.wavelength > 1*u.micron, :]
        r[:, np.abs(r.time) < 1*u.hour]
        r[r.wavelength > 1*u.micron, np.abs(r.time) < 1*u.hour]
        ```

        Parameters
        ----------
        key : tuple
            The (wavelength, time) slices, indices, or masks.
        """

        i_wavelength, i_time = key

        # create a copy
        new = self._create_copy()

        # do indexing of wavelike
        for w in self.wavelike:
            new.wavelike[w] = self.wavelike[w][i_wavelength]

        # do indexing of timelike
        for t in self.timelike:
            new.timelike[t] = self.timelike[t][i_time]

        # do indexing of fluxlike
        for f in self.fluxlike:
            # (indexing step by step seems more stable)
            if self.fluxlike[f] is None:
                continue
            temporary = self.fluxlike[f][i_wavelength, :]
            new.fluxlike[f] = temporary[:, i_time]

        # finalize the new rainbow
        new._validate_core_dictionaries()
        new._guess_wscale()

        return new

    def __repr__(self):
        """
        How should this object be represented as a string?
        """
        n = self.__class__.__name__.replace("Rainbow", "🌈")
        return f"<{n}({self.nwave}w, {self.ntime}t)>"

    # import the basic operations for Rainbows
    from .operations import __add__, __sub__, __mul__, __truediv__, __eq__

    # import other axtions that return other Rainbows
    from .actions import (
        normalize,
        bin,
        bin_in_time,
        bin_in_wavelength,
        get_spectrum,
        get_spectral_resolution,
        plot_spectral_resolution,
    )

    # import visualizations that can act on Rainbows
    from .visualizations import (
        imshow,
        plot,
        _setup_animate_lightcurves,
        animate_lightcurves,
        _setup_animate_spectra,
        animate_spectra,
        _setup_animated_scatter,
        _setup_wavelength_colors,
        _make_sure_cmap_is_defined,
        get_wavelength_color,
    )



from ..rainbows import *
from .setup_tests import *


def test_simulated_basics():

    a = SimulatedRainbow()
    b = SimulatedRainbow(tlim=[-1, 1] * u.hour, dt=1 * u.minute)
    c = SimulatedRainbow(time=np.linspace(-1, 1) * u.hour)
    d = SimulatedRainbow(wlim=[0.1, 1] * u.micron, R=100)
    assert d.wscale == "log"
    e = SimulatedRainbow(wlim=[0.1, 1] * u.micron, dw=1 * u.nm)
    assert e.wscale == "linear"
    f = SimulatedRainbow(wavelength=np.logspace(0, 1) * u.micron)


def test_star_flux():
    g = SimulatedRainbow(
        wavelength=np.logspace(0, 1) * u.micron, star_flux=np.logspace(0, 1)
    )


def test_inject_transit():
    h = SimulatedRainbow(wavelength=np.logspace(0, 1) * u.micron)
    h.inject_transit(planet_params={"per": 0.1})
    h.inject_transit(planet_radius=np.zeros(50) + 0.1)

    i = SimulatedRainbow(signal_to_noise=1000, dt=2 * u.minute)
    fi, ax = plt.subplots(3, 1, sharex=True, figsize=(8, 8))
    i.imshow(ax=ax[0], vmin=0.975, vmax=1.005)
    plt.xlabel("")
    i.inject_transit(
        planet_params=dict(per=3), planet_radius=np.random.normal(0.1, 0.01, i.nwave)
    ).imshow(ax=ax[1], vmin=0.975, vmax=1.005)
    # test the limb darkening
    i.inject_transit(
        planet_params={
            "per": 3,
            "limb_dark": "quadratic",
            "u": np.transpose(
                [np.linspace(1.0, 0.0, i.nwave), np.linspace(0.5, 0.0, i.nwave)]
            ),
        },
    ).imshow(ax=ax[2], vmin=0.975, vmax=1.005)
    plt.savefig(os.path.join(test_directory, "transit-injection-demonstration.pdf"))



from ...imports import *

__all__ = ["normalize"]


def normalize(self, wavelength=True, time=False):
    """
    Normalize by dividing through by the median spectrum and/or lightcurve.

    Parameters
    ----------
    wavelength : bool
        Should we divide by the median spectrum?

    time : bool
        Should we divide by the median light curve?

    Returns
    -------
    normalized : MultiRainbow
        The normalized MultiRainbow.
    """

    # TODO, think about more careful treatment of uncertainties + good/bad data
    new = self._create_copy()

    # (ignore nan warnings)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if time:
            new = new / np.nanmedian(self.flux, axis=self.waveaxis)

        if wavelength:
            new = new / np.nanmedian(self.flux, axis=self.timeaxis)

    return new



"""
If something inherits from Talker, then we can print
text to the terminal in a relatively standard way.
"""
import textwrap, numpy as np, pprint
import sys

if sys.version_info[0] < 3:
    input = raw_input

shortcuts = None

# TODO this whole thing is overly complicated. simplify?


class Talker(object):
    """
    Objects the inherit from Talker have...
    "mute" and "pithy" attributes,
    a report('uh-oh!') method that prints when unmuted,
    a speak('yo!') method that prints only when unmuted and unpithy,
    and an input("what's up?") method that takes input from the prompt.
    """

    _mute = False
    _pithy = False
    line = np.inf

    @property
    def _nametag(self):
        return self.__class__.__name__.lower()

    def speak(self, string="", level=0, progress=False):
        """
        If verbose=True and terse=False,
        this will print to terminal.
        Otherwise, it won't.
        """
        if self._pithy == False:
            self.report(string=string, level=level, progress=progress)

    def warning(self, string="", level=0):
        """
        If verbose=True and terse=False,
        this will print to terminal.
        Otherwise, it won't.
        """
        self.report(string, level, prelude=":-| ")

    def input(self, string="", level=0, prompt="(please respond) "):
        """
        If verbose=True and terse=False,
        this will print to terminal.
        Otherwise, it won't.
        """
        self.report(string, level)
        return input("{0}".format(self._prefix + prompt))

    def report(self, string="", level=0, prelude="", progress=False, abbreviate=True):
        """
        If verbose=True,
        this will print to terminal.
        Otherwise, it won't.
        """
        if self._mute == False:
            self._prefix = prelude + "{spacing}{name}: ".format(
                name=self._nametag, spacing=" " * level
            )
            self._prefix = "{0:>16}".format(self._prefix)
            equalspaces = " " * len(self._prefix)
            toprint = string + ""
            if abbreviate:
                if shortcuts is not None:
                    for k in shortcuts.keys():
                        toprint = toprint.replace(k, shortcuts[k])

            if progress:
                print(
                    "\r" + self._prefix + toprint.replace("\n", "\n" + equalspaces),
                )
            else:
                print(self._prefix + toprint.replace("\n", "\n" + equalspaces))

    def summarize(self):
        """
        Print a summary of the contents of this object.
        """

        self.speak("Here is a brief summary of {}.".format(self._nametag))
        s = "\n" + pprint.pformat(self.__dict__)
        print(s.replace("\n", "\n" + " " * (len(self._prefix) + 1)) + "\n")



from ..rainbows import *
from .setup_tests import *


def test_rainbow_basics():

    nw, nt = 24, 48
    r = Rainbow(
        wavelength=np.linspace(1, 2, nw) * u.micron,
        time=np.linspace(-1, 1, nt) * u.hour,
        flux=np.ones((nw, nt)),
    )

    # test the basic sizes and shapes
    assert r.shape == (nw, nt)
    assert r.nwave == nw
    assert r.ntime == nt
    assert r.nflux == nw * nt

    # testing slicing, indexing, masking
    assert r[:, :].shape == (nw, nt)
    assert r[::2, ::3].shape == (nw / 2, nt / 3)
    assert r[2:4, :].shape == (2, nt)
    assert r[np.arange(2, 4), :].shape == (2, nt)
    assert r[r.wavelength < 1.5 * u.micron, :].shape == (nw / 2, nt)
    assert r[:, r.time < 0 * u.day].shape == (nw, nt / 2)

    assert r[r.wavelength < 1.5 * u.micron, r.time < 0 * u.day].shape == (
        nw / 2,
        nt / 2,
    )
